@@ -501,12 +501,19 @@ load_thumbnail_info (const char *path)
   if (pixbuf == NULL)
     return NULL;
   
+  thumb_uri = gdk_pixbuf_get_option (pixbuf, "tEXt::Thumb::URI");
+  thumb_mtime_str = gdk_pixbuf_get_option (pixbuf, "tEXt::Thumb::MTime");
+
+  if (thumb_uri == NULL ||
+      thumb_mtime_str == NULL)
+    {
+      g_object_unref (pixbuf);
+      return NULL;
+    }
+
   info = g_new0 (struct ThumbnailInfo, 1);
   
-  thumb_uri = gdk_pixbuf_get_option (pixbuf, "tEXt::Thumb::URI");
   info->uri = g_strdup (thumb_uri);
-  
-  thumb_mtime_str = gdk_pixbuf_get_option (pixbuf, "tEXt::Thumb::MTime");
   info->mtime = atol (thumb_mtime_str);
 
   g_object_unref (pixbuf);
@@ -1297,6 +1304,9 @@ gnome_thumbnail_has_uri (GdkPixbuf          *pixbuf,
   const char *thumb_uri;
   
   thumb_uri = gdk_pixbuf_get_option (pixbuf, "tEXt::Thumb::URI");
+  if (!thumb_uri)
+    return FALSE;
+
   return strcmp (uri, thumb_uri) == 0;
 }
 
@@ -1320,10 +1330,14 @@ gnome_thumbnail_is_valid (GdkPixbuf          *pixbuf,
   time_t thumb_mtime;
   
   thumb_uri = gdk_pixbuf_get_option (pixbuf, "tEXt::Thumb::URI");
+  if (!thumb_uri)
+    return FALSE;
   if (strcmp (uri, thumb_uri) != 0)
     return FALSE;
   
   thumb_mtime_str = gdk_pixbuf_get_option (pixbuf, "tEXt::Thumb::MTime");
+  if (!thumb_mtime_str)
+    return FALSE;
   thumb_mtime = atol (thumb_mtime_str);
   if (mtime != thumb_mtime)
     return FALSE;
